package com.thinkaurelius.titan.graphdb;

import com.google.common.collect.Iterables;
import com.thinkaurelius.titan.core.TitanKey;
import com.thinkaurelius.titan.core.TitanLabel;
import com.thinkaurelius.titan.core.attribute.Cmp;
import com.thinkaurelius.titan.core.attribute.Geo;
import com.thinkaurelius.titan.core.attribute.Geoshape;
import com.thinkaurelius.titan.core.attribute.Text;
import com.tinkerpop.blueprints.Direction;
import com.tinkerpop.blueprints.Edge;
import com.tinkerpop.blueprints.Vertex;
import org.apache.commons.configuration.Configuration;
import org.junit.Test;

import static org.junit.Assert.assertEquals;

/**
 * @author Matthias Broecheler (me@matthiasb.com)
 */

public abstract class TitanIndexTest extends TitanGraphTestCommon  {

    public static final String INDEX = "index";

    public final boolean supportsGeoPoint;
    public final boolean supportsNumeric;
    public final boolean supportsText;

    protected TitanIndexTest(Configuration config, boolean supportsGeoPoint, boolean supportsNumeric, boolean supportsText) {
        super(config);
        this.supportsGeoPoint = supportsGeoPoint;
        this.supportsNumeric = supportsNumeric;
        this.supportsText = supportsText;
    }

    @Test
    public void testOpenClose() {
    }

    @Test
    public void testSimpleUpdate() {
        TitanKey text = tx.makeType().name("name").unique(Direction.OUT)
                .indexed(INDEX,Vertex.class).indexed(INDEX, Edge.class).dataType(String.class).makePropertyKey();
        Vertex v = tx.addVertex();
        v.setProperty("name","Marko Rodriguez");
        clopen();
        Iterable<Vertex> vs = tx.query().has("name",Text.CONTAINS,"marko").vertices();
        assertEquals(1,Iterables.size(vs));
        v = vs.iterator().next();
        v.setProperty("name","Marko");
        clopen();
        vs = tx.query().has("name",Text.CONTAINS,"marko").vertices();
        assertEquals(1,Iterables.size(vs));
        v = vs.iterator().next();

    }

    @Test
    public void testIndexing() {
        TitanKey text = tx.makeType().name("text").unique(Direction.OUT)
                .indexed(INDEX,Vertex.class).indexed(INDEX, Edge.class).dataType(String.class).makePropertyKey();
        TitanKey location = tx.makeType().name("location").unique(Direction.OUT)
                .indexed(INDEX,Vertex.class).indexed(INDEX, Edge.class).dataType(Geoshape.class).makePropertyKey();
        TitanKey time = tx.makeType().name("time").unique(Direction.OUT)
                .indexed(INDEX, Vertex.class).indexed(INDEX, Edge.class).dataType(Long.class).makePropertyKey();
        TitanKey id = tx.makeType().name("uid").unique(Direction.OUT).unique(Direction.IN)
                .indexed(Vertex.class).dataType(Integer.class).makePropertyKey();
        TitanLabel knows = tx.makeType().name("knows").primaryKey(time).signature(location).makeEdgeLabel();

        clopen();
        String[] words = {"world","aurelius","titan","graph"};
        double distance, offset;
        int numV = 100; final int originalNumV = numV;
        for (int i=0;i<numV;i++) {
            Vertex v = tx.addVertex();
            v.setProperty("uid",i);
            v.setProperty("text", "Vertex " + words[i%words.length]);
            v.setProperty("time",i);
            offset = (i%2==0?1:-1)*(i*50.0/numV);
            v.setProperty("location",Geoshape.point(0.0+offset,0.0+offset));

            Edge e = v.addEdge("knows",tx.getVertex("uid",Math.max(0,i-1)));
            e.setProperty("text", "Vertex " + words[i%words.length]);
            e.setProperty("time",i);
            e.setProperty("location", Geoshape.point(0.0 + offset, 0.0 + offset));
        }

        for (int i=0;i<words.length;i++) {
            assertEquals(numV/words.length, Iterables.size(tx.query().has("text", Text.CONTAINS, words[i]).vertices()));
            assertEquals(numV/words.length, Iterables.size(tx.query().has("text", Text.CONTAINS, words[i]).edges()));
        }

        for (int i=0;i<numV/2;i+=numV/10) {
            assertEquals(i, Iterables.size(tx.query().has("time", Cmp.GREATER_THAN_EQUAL, i).has("time", Cmp.LESS_THAN, i + i).vertices()));
            assertEquals(i, Iterables.size(tx.query().has("time", Cmp.GREATER_THAN_EQUAL, i).has("time", Cmp.LESS_THAN, i + i).edges()));
        }

        for (int i=0;i<numV;i+=10) {
            offset = (i*50.0/originalNumV);
            distance = Geoshape.point(0.0,0.0).getPoint().distance(Geoshape.point(offset,offset).getPoint())+20;
            assertEquals(i+1, Iterables.size(tx.query().has("location", Geo.WITHIN, Geoshape.circle(0.0, 0.0, distance)).vertices()));
            assertEquals(i+1, Iterables.size(tx.query().has("location", Geo.WITHIN, Geoshape.circle(0.0, 0.0, distance)).edges()));
        }

        assertEquals(5, Iterables.size(tx.query().has("time", Cmp.GREATER_THAN_EQUAL, 10).has("time", Cmp.LESS_THAN, 30).has("text", Text.CONTAINS, words[0]).vertices()));
        offset = (19*50.0/originalNumV);
        distance = Geoshape.point(0.0,0.0).getPoint().distance(Geoshape.point(offset,offset).getPoint())+20;
        assertEquals(5, Iterables.size(tx.query().has("location", Geo.INTERSECT, Geoshape.circle(0.0, 0.0, distance)).has("text", Text.CONTAINS, words[0]).vertices()));

        assertEquals(numV, Iterables.size(tx.getVertices()));
        assertEquals(numV, Iterables.size(tx.getEdges()));

        clopen();

        //Copied from above
        for (int i=0;i<words.length;i++) {
            assertEquals(numV/words.length, Iterables.size(tx.query().has("text", Text.CONTAINS, words[i]).vertices()));
            assertEquals(numV/words.length, Iterables.size(tx.query().has("text", Text.CONTAINS, words[i]).edges()));
        }

        for (int i=0;i<numV/2;i+=numV/10) {
            assertEquals(i, Iterables.size(tx.query().has("time", Cmp.GREATER_THAN_EQUAL, i).has("time", Cmp.LESS_THAN, i + i).vertices()));
            assertEquals(i, Iterables.size(tx.query().has("time", Cmp.GREATER_THAN_EQUAL, i).has("time", Cmp.LESS_THAN, i + i).edges()));
        }

        for (int i=0;i<numV;i+=10) {
            offset = (i*50.0/originalNumV);
            distance = Geoshape.point(0.0,0.0).getPoint().distance(Geoshape.point(offset,offset).getPoint())+20;
            assertEquals(i+1, Iterables.size(tx.query().has("location", Geo.WITHIN, Geoshape.circle(0.0, 0.0, distance)).vertices()));
            assertEquals(i+1, Iterables.size(tx.query().has("location", Geo.WITHIN, Geoshape.circle(0.0, 0.0, distance)).edges()));
        }

        assertEquals(5, Iterables.size(tx.query().has("time", Cmp.GREATER_THAN_EQUAL, 10).has("time", Cmp.LESS_THAN, 30).has("text", Text.CONTAINS, words[0]).vertices()));
        offset = (19*50.0/originalNumV);
        distance = Geoshape.point(0.0,0.0).getPoint().distance(Geoshape.point(offset,offset).getPoint())+20;
        assertEquals(5, Iterables.size(tx.query().has("location", Geo.INTERSECT, Geoshape.circle(0.0, 0.0, distance)).has("text", Text.CONTAINS, words[0]).vertices()));

        assertEquals(numV, Iterables.size(tx.getVertices()));
        assertEquals(numV, Iterables.size(tx.getEdges()));

        newTx();

        int numDelete = 12;
        for (int i=numV-numDelete; i<numV;i++) {
            tx.getVertex("uid",i).remove();
        }

        numV = numV-numDelete;

        //Copied from above
        for (int i=0;i<words.length;i++) {
            assertEquals(numV/words.length, Iterables.size(tx.query().has("text", Text.CONTAINS, words[i]).vertices()));
            assertEquals(numV/words.length, Iterables.size(tx.query().has("text", Text.CONTAINS, words[i]).edges()));
        }

        for (int i=0;i<numV/2;i+=numV/10) {
            assertEquals(i, Iterables.size(tx.query().has("time", Cmp.GREATER_THAN_EQUAL, i).has("time", Cmp.LESS_THAN, i + i).vertices()));
            assertEquals(i, Iterables.size(tx.query().has("time", Cmp.GREATER_THAN_EQUAL, i).has("time", Cmp.LESS_THAN, i + i).edges()));
        }

        for (int i=0;i<numV;i+=10) {
            offset = (i*50.0/originalNumV);
            distance = Geoshape.point(0.0,0.0).getPoint().distance(Geoshape.point(offset,offset).getPoint())+20;
            assertEquals(i+1, Iterables.size(tx.query().has("location", Geo.WITHIN, Geoshape.circle(0.0, 0.0, distance)).vertices()));
            assertEquals(i+1, Iterables.size(tx.query().has("location", Geo.WITHIN, Geoshape.circle(0.0, 0.0, distance)).edges()));
        }

        assertEquals(5, Iterables.size(tx.query().has("time", Cmp.GREATER_THAN_EQUAL, 10).has("time", Cmp.LESS_THAN, 30).has("text", Text.CONTAINS, words[0]).vertices()));
        offset = (19*50.0/originalNumV);
        distance = Geoshape.point(0.0,0.0).getPoint().distance(Geoshape.point(offset,offset).getPoint())+20;
        assertEquals(5, Iterables.size(tx.query().has("location", Geo.INTERSECT, Geoshape.circle(0.0, 0.0, distance)).has("text", Text.CONTAINS, words[0]).vertices()));

        assertEquals(numV, Iterables.size(tx.getVertices()));
        assertEquals(numV, Iterables.size(tx.getEdges()));

    }

}
